/*
 * kgraft_patch_cve-2015-7872
 *
 * Fix for CVE-2015-7872, bsc#951542
 *
 *  Upstream commits:
 *  94c4554ba07adbdde396748ee7ae01e86cf2d8d7
 *  911b79cde95c7da0ec02f48105358a36636b7a71
 *  f05819df10d7b09f6d1eb6f8534a8f68e5a4fe61
 *  a3a8784454692dd72e5d5d34dcdab17b4420e74c
 *
 *  Copyright (c) 2015 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/cred.h>
#include <linux/sched.h>
#include <linux/key.h>
#include <linux/key-type.h>
#include <linux/slab.h>

#include "local.h"

static struct key_user *(*kgr_key_user_lookup)(kuid_t uid);
static void (*kgr_construct_get_dest_keyring)(struct key **_dest_keyring);
static int (*kgr__key_link_begin)(struct key *keyring, const struct keyring_index_key *index_key, struct assoc_array_edit **_edit);
static struct mutex *kgr_key_construction_mutex;
static key_ref_t (*kgr_search_process_keyrings)(struct keyring_search_context *ctx);
static void (*kgr__key_link)(struct key *key, struct assoc_array_edit **_edit);
static void (*kgr__key_link_end)(struct key *keyring, const struct keyring_index_key *index_key, struct assoc_array_edit *edit);
static int (*kgr__key_link_check_live_key)(struct key *keyring, struct key *key);
static void (*kgr_key_user_put)(struct key_user *user);
static struct key *(*kgr_request_key_auth_new)(struct key *target, const void *callout_info, size_t callout_len, struct key *dest_keyring);
static int (*kgr_call_sbin_request_key)(struct key_construction *cons, const char *op, void *aux);
static void (*kgr_security_key_free)(struct key *key);
static struct kmem_cache **kgr_key_jar;
static unsigned long *kgr_key_gc_flags;
static unsigned *kgr_key_gc_delay;
static spinlock_t *kgr_key_serial_lock;
static struct rb_root *kgr_key_serial_tree;
static struct key_type **kgr_key_gc_dead_keytype;
static void (*kgr_key_schedule_gc)(time_t gc_at);
static struct work_struct *kgr_key_gc_work;
static struct key_type *kgr_key_type_dead;
static int (*kgr_key_gc_keyring_func)(const void *object, void *iterator_data);
static void (*kgr_keyring_gc)(struct key *keyring, time_t limit);
static int (*kgr_assoc_array_iterate)(const struct assoc_array *array, int (*iterator)(const void *object, void *iterator_data), void *iterator_data);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "key_user_lookup", (void *)&kgr_key_user_lookup },
	{ "construct_get_dest_keyring", (void *)&kgr_construct_get_dest_keyring },
	{ "__key_link_begin", (void *)&kgr__key_link_begin },
	{ "key_construction_mutex", (void *)&kgr_key_construction_mutex },
	{ "search_process_keyrings", (void *)&kgr_search_process_keyrings },
	{ "__key_link", (void *)&kgr__key_link },
	{ "__key_link_end", (void *)&kgr__key_link_end },
	{ "__key_link_check_live_key", (void *)&kgr__key_link_check_live_key },
	{ "key_user_put", (void *)&kgr_key_user_put },
	{ "request_key_auth_new", (void *)&kgr_request_key_auth_new },
	{ "call_sbin_request_key", (void *)&kgr_call_sbin_request_key },
	{ "security_key_free", (void *)&kgr_security_key_free },
	{ "key_jar", (void *)&kgr_key_jar },
	{ "key_gc_flags", (void *)&kgr_key_gc_flags },
	{ "key_gc_delay", (void *)&kgr_key_gc_delay },
	{ "key_serial_lock", (void *)&kgr_key_serial_lock },
	{ "key_serial_tree", (void *)&kgr_key_serial_tree },
	{ "key_gc_dead_keytype", (void *)&kgr_key_gc_dead_keytype },
	{ "key_schedule_gc", (void *)&kgr_key_schedule_gc },
	{ "key_gc_work", (void *)&kgr_key_gc_work },
	{ "key_type_dead", (void *)&kgr_key_type_dead },
	{ "key_gc_keyring_func", (void *)&kgr_key_gc_keyring_func },
	{ "keyring_gc", (void *)&kgr_keyring_gc },
	{ "assoc_array_iterate", (void *)&kgr_assoc_array_iterate },
};

#define key_negative_timeout	60	/* default timeout on a negative key's existence */

#define KEY_GC_KEY_EXPIRED	0	/* A key expired and needs unlinking */
#define KEY_GC_REAP_KEYTYPE	1	/* A keytype is being unregistered */
#define KEY_GC_REAPING_KEYTYPE	2	/* Cleared when keytype reaped */

static int kgr_construct_alloc_key(struct keyring_search_context *ctx,
			       struct key *dest_keyring,
			       unsigned long flags,
			       struct key_user *user,
			       struct key **_key)
{
	struct assoc_array_edit *edit;
	struct key *key;
	key_perm_t perm;
	key_ref_t key_ref;
	int ret;

	kenter("%s,%s,,,",
	       ctx->index_key.type->name, ctx->index_key.description);

	*_key = NULL;
	mutex_lock(&user->cons_lock);

	perm = KEY_POS_VIEW | KEY_POS_SEARCH | KEY_POS_LINK | KEY_POS_SETATTR;
	perm |= KEY_USR_VIEW;
	if (ctx->index_key.type->read)
		perm |= KEY_POS_READ;
	if (ctx->index_key.type == &key_type_keyring ||
	    ctx->index_key.type->update)
		perm |= KEY_POS_WRITE;

	key = key_alloc(ctx->index_key.type, ctx->index_key.description,
			ctx->cred->fsuid, ctx->cred->fsgid, ctx->cred,
			perm, flags);
	if (IS_ERR(key))
		goto alloc_failed;

	set_bit(KEY_FLAG_USER_CONSTRUCT, &key->flags);

	if (dest_keyring) {
		ret = kgr__key_link_begin(dest_keyring, &ctx->index_key, &edit);
		if (ret < 0)
			goto link_prealloc_failed;
	}

	/* attach the key to the destination keyring under lock, but we do need
	 * to do another check just in case someone beat us to it whilst we
	 * waited for locks */
	mutex_lock(kgr_key_construction_mutex);

	key_ref = kgr_search_process_keyrings(ctx);
	if (!IS_ERR(key_ref))
		goto key_already_present;

	if (dest_keyring)
		kgr__key_link(key, &edit);

	mutex_unlock(kgr_key_construction_mutex);
	if (dest_keyring)
		kgr__key_link_end(dest_keyring, &ctx->index_key, edit);
	mutex_unlock(&user->cons_lock);
	*_key = key;
	kleave(" = 0 [%d]", key_serial(key));
	return 0;

	/* the key is now present - we tell the caller that we found it by
	 * returning -EINPROGRESS  */
key_already_present:
	key_put(key);
	mutex_unlock(kgr_key_construction_mutex);
	key = key_ref_to_ptr(key_ref);
	if (dest_keyring) {
		ret = kgr__key_link_check_live_key(dest_keyring, key);
		if (ret == 0)
			kgr__key_link(key, &edit);
		kgr__key_link_end(dest_keyring, &ctx->index_key, edit);
		if (ret < 0)
			goto link_check_failed;
	}
	mutex_unlock(&user->cons_lock);
	*_key = key;
	kleave(" = -EINPROGRESS [%d]", key_serial(key));
	return -EINPROGRESS;

link_check_failed:
	mutex_unlock(&user->cons_lock);
	key_put(key);
	kleave(" = %d [linkcheck]", ret);
	return ret;

link_prealloc_failed:
	mutex_unlock(&user->cons_lock);
	kleave(" = %d [prelink]", ret);
	return ret;

alloc_failed:
	mutex_unlock(&user->cons_lock);
	kleave(" = %ld", PTR_ERR(key));
	return PTR_ERR(key);
}

static int kgr_construct_key(struct key *key, const void *callout_info,
			 size_t callout_len, void *aux,
			 struct key *dest_keyring)
{
	struct key_construction *cons;
	request_key_actor_t actor;
	struct key *authkey;
	int ret;

	kenter("%d,%p,%zu,%p", key->serial, callout_info, callout_len, aux);

	cons = kmalloc(sizeof(*cons), GFP_KERNEL);
	if (!cons)
		return -ENOMEM;

	/* allocate an authorisation key */
	authkey = kgr_request_key_auth_new(key, callout_info, callout_len,
				       dest_keyring);
	if (IS_ERR(authkey)) {
		kfree(cons);
		ret = PTR_ERR(authkey);
		authkey = NULL;
	} else {
		cons->authkey = key_get(authkey);
		cons->key = key_get(key);

		/* make the call */
		actor = kgr_call_sbin_request_key;
		if (key->type->request_key)
			actor = key->type->request_key;

		ret = actor(cons, "create", aux);

		/* check that the actor called complete_request_key() prior to
		 * returning an error */
		WARN_ON(ret < 0 &&
			!test_bit(KEY_FLAG_REVOKED, &authkey->flags));
		key_put(authkey);
	}

	kleave(" = %d", ret);
	return ret;
}

/* Patched function, but inlined to request_key_and_link() */
static struct key *kgr_construct_key_and_link(struct keyring_search_context *ctx,
	const char *callout_info, size_t callout_len, void *aux,
	struct key *dest_keyring, unsigned long flags)
{
	struct key_user *user;
	struct key *key;
	int ret;

	kenter("");

	/* Fix CVE-2015-7872 */
	if (ctx->index_key.type == &key_type_keyring)
		return ERR_PTR(-EPERM);

	user = kgr_key_user_lookup(current_fsuid());
	if (!user)
		return ERR_PTR(-ENOMEM);

	kgr_construct_get_dest_keyring(&dest_keyring);

	ret = kgr_construct_alloc_key(ctx, dest_keyring, flags, user, &key);
	kgr_key_user_put(user);

	if (ret == 0) {
		ret = kgr_construct_key(key, callout_info, callout_len, aux,
				    dest_keyring);
		if (ret < 0) {
			kdebug("cons failed");
			goto construction_failed;
		}
	} else if (ret == -EINPROGRESS) {
		ret = 0;
	} else {
		goto couldnt_alloc_key;
	}

	key_put(dest_keyring);
	kleave(" = key %d", key_serial(key));
	return key;

construction_failed:
	key_negate_and_link(key, key_negative_timeout, NULL, NULL);
	key_put(key);
couldnt_alloc_key:
	key_put(dest_keyring);
	kleave(" = %d", ret);
	return ERR_PTR(ret);
}

struct key *kgr_request_key_and_link(struct key_type *type, const char *description,
	const void *callout_info, size_t callout_len, void *aux,
	struct key *dest_keyring, unsigned long flags)
{
	struct keyring_search_context ctx = {
		.index_key.type		= type,
		.index_key.description	= description,
		.cred			= current_cred(),
		.match			= type->match,
		.match_data		= description,
		.flags			= KEYRING_SEARCH_LOOKUP_DIRECT,
	};
	struct key *key;
	key_ref_t key_ref;
	int ret;

	kenter("%s,%s,%p,%zu,%p,%p,%lx",
	       ctx.index_key.type->name, ctx.index_key.description,
	       callout_info, callout_len, aux, dest_keyring, flags);

	/* search all the process keyrings for a key */
	key_ref = kgr_search_process_keyrings(&ctx);

	if (!IS_ERR(key_ref)) {
		key = key_ref_to_ptr(key_ref);
		if (dest_keyring) {
			kgr_construct_get_dest_keyring(&dest_keyring);
			ret = key_link(dest_keyring, key);
			key_put(dest_keyring);
			if (ret < 0) {
				key_put(key);
				key = ERR_PTR(ret);
				goto error;
			}
		}
	} else if (PTR_ERR(key_ref) != -EAGAIN) {
		key = ERR_CAST(key_ref);
	} else  {
		/* the search failed, but the keyrings were searchable, so we
		 * should consult userspace if we can */
		key = ERR_PTR(-ENOKEY);
		if (!callout_info)
			goto error;

		key = kgr_construct_key_and_link(&ctx, callout_info, callout_len,
					     aux, dest_keyring, flags);
	}

error:
	kleave(" = %p", key);
	return key;
}

/* Patched function but optimized -> need to patch key_garbage_collector */
static noinline void kgr_key_gc_unused_keys(struct list_head *keys)
{
	while (!list_empty(keys)) {
		struct key *key =
			list_entry(keys->next, struct key, graveyard_link);
		list_del(&key->graveyard_link);

		kdebug("- %u", key->serial);
		/* KGR - KEY_DEBUGGING is off by default */
		/* key_check(key); */

		/* Fix CVE-2015-7872 */
		/* Throw away the key data if the key is instantiated */
		if (test_bit(KEY_FLAG_INSTANTIATED, &key->flags) &&
		    !test_bit(KEY_FLAG_NEGATIVE, &key->flags) &&
		    key->type->destroy)
			key->type->destroy(key);

		kgr_security_key_free(key);

		/* deal with the user's key tracking and quota */
		if (test_bit(KEY_FLAG_IN_QUOTA, &key->flags)) {
			spin_lock(&key->user->lock);
			key->user->qnkeys--;
			key->user->qnbytes -= key->quotalen;
			spin_unlock(&key->user->lock);
		}

		atomic_dec(&key->user->nkeys);
		if (test_bit(KEY_FLAG_INSTANTIATED, &key->flags))
			atomic_dec(&key->user->nikeys);

		/* Fix CVE-2014-9529 */
		kgr_key_user_put(key->user);

		kfree(key->description);

#ifdef KEY_DEBUGGING
		key->magic = KEY_DEBUG_MAGIC_X;
#endif
		kmem_cache_free(*kgr_key_jar, key);
	}
}

static void kgr_key_gc_keyring(struct key *keyring, time_t limit)
{
	int result;

	kenter("%x{%s}", keyring->serial, keyring->description ?: "");

	if (keyring->flags & ((1 << KEY_FLAG_INVALIDATED) |
			      (1 << KEY_FLAG_REVOKED)))
		goto dont_gc;

	/* scan the keyring looking for dead keys */
	rcu_read_lock();
	result = kgr_assoc_array_iterate(&keyring->keys,
					 kgr_key_gc_keyring_func, &limit);
	rcu_read_unlock();
	if (result == true)
		goto do_gc;

dont_gc:
	kleave(" [no gc]");
	return;

do_gc:
	kgr_keyring_gc(keyring, limit);
	kleave(" [gc]");
}

void kgr_key_garbage_collector(struct work_struct *work)
{
	static LIST_HEAD(graveyard);
	static u8 gc_state;		/* Internal persistent state */
#define KEY_GC_REAP_AGAIN	0x01	/* - Need another cycle */
#define KEY_GC_REAPING_LINKS	0x02	/* - We need to reap links */
#define KEY_GC_SET_TIMER	0x04	/* - We need to restart the timer */
#define KEY_GC_REAPING_DEAD_1	0x10	/* - We need to mark dead keys */
#define KEY_GC_REAPING_DEAD_2	0x20	/* - We need to reap dead key links */
#define KEY_GC_REAPING_DEAD_3	0x40	/* - We need to reap dead keys */
#define KEY_GC_FOUND_DEAD_KEY	0x80	/* - We found at least one dead key */

	struct rb_node *cursor;
	struct key *key;
	time_t new_timer, limit;

	kenter("[%lx,%x]", *kgr_key_gc_flags, gc_state);

	limit = current_kernel_time().tv_sec;
	if (limit > *kgr_key_gc_delay)
		limit -= *kgr_key_gc_delay;
	else
		limit = *kgr_key_gc_delay;

	/* Work out what we're going to be doing in this pass */
	gc_state &= KEY_GC_REAPING_DEAD_1 | KEY_GC_REAPING_DEAD_2;
	gc_state <<= 1;
	if (test_and_clear_bit(KEY_GC_KEY_EXPIRED, kgr_key_gc_flags))
		gc_state |= KEY_GC_REAPING_LINKS | KEY_GC_SET_TIMER;

	if (test_and_clear_bit(KEY_GC_REAP_KEYTYPE, kgr_key_gc_flags))
		gc_state |= KEY_GC_REAPING_DEAD_1;
	kdebug("new pass %x", gc_state);

	new_timer = LONG_MAX;

	/* As only this function is permitted to remove things from the key
	 * serial tree, if cursor is non-NULL then it will always point to a
	 * valid node in the tree - even if lock got dropped.
	 */
	spin_lock(kgr_key_serial_lock);
	cursor = rb_first(kgr_key_serial_tree);

continue_scanning:
	while (cursor) {
		key = rb_entry(cursor, struct key, serial_node);
		cursor = rb_next(cursor);

		if (atomic_read(&key->usage) == 0)
			goto found_unreferenced_key;

		if (unlikely(gc_state & KEY_GC_REAPING_DEAD_1)) {
			if (key->type == *kgr_key_gc_dead_keytype) {
				gc_state |= KEY_GC_FOUND_DEAD_KEY;
				set_bit(KEY_FLAG_DEAD, &key->flags);
				key->perm = 0;
				goto skip_dead_key;
			}
		}

		if (gc_state & KEY_GC_SET_TIMER) {
			if (key->expiry > limit && key->expiry < new_timer) {
				kdebug("will expire %x in %ld",
				       key_serial(key), key->expiry - limit);
				new_timer = key->expiry;
			}
		}

		if (unlikely(gc_state & KEY_GC_REAPING_DEAD_2))
			if (key->type == *kgr_key_gc_dead_keytype)
				gc_state |= KEY_GC_FOUND_DEAD_KEY;

		if ((gc_state & KEY_GC_REAPING_LINKS) ||
		    unlikely(gc_state & KEY_GC_REAPING_DEAD_2)) {
			if (key->type == &key_type_keyring)
				goto found_keyring;
		}

		if (unlikely(gc_state & KEY_GC_REAPING_DEAD_3))
			if (key->type == *kgr_key_gc_dead_keytype)
				goto destroy_dead_key;

	skip_dead_key:
		if (spin_is_contended(kgr_key_serial_lock) || need_resched())
			goto contended;
	}

contended:
	spin_unlock(kgr_key_serial_lock);

maybe_resched:
	if (cursor) {
		cond_resched();
		spin_lock(kgr_key_serial_lock);
		goto continue_scanning;
	}

	/* We've completed the pass.  Set the timer if we need to and queue a
	 * new cycle if necessary.  We keep executing cycles until we find one
	 * where we didn't reap any keys.
	 */
	kdebug("pass complete");

	if (gc_state & KEY_GC_SET_TIMER && new_timer != (time_t)LONG_MAX) {
		new_timer += *kgr_key_gc_delay;
		kgr_key_schedule_gc(new_timer);
	}

	if (unlikely(gc_state & KEY_GC_REAPING_DEAD_2) ||
	    !list_empty(&graveyard)) {
		/* Make sure that all pending keyring payload destructions are
		 * fulfilled and that people aren't now looking at dead or
		 * dying keys that they don't have a reference upon or a link
		 * to.
		 */
		kdebug("gc sync");
		synchronize_rcu();
	}

	if (!list_empty(&graveyard)) {
		kdebug("gc keys");
		kgr_key_gc_unused_keys(&graveyard);
	}

	if (unlikely(gc_state & (KEY_GC_REAPING_DEAD_1 |
				 KEY_GC_REAPING_DEAD_2))) {
		if (!(gc_state & KEY_GC_FOUND_DEAD_KEY)) {
			/* No remaining dead keys: short circuit the remaining
			 * keytype reap cycles.
			 */
			kdebug("dead short");
			gc_state &= ~(KEY_GC_REAPING_DEAD_1 | KEY_GC_REAPING_DEAD_2);
			gc_state |= KEY_GC_REAPING_DEAD_3;
		} else {
			gc_state |= KEY_GC_REAP_AGAIN;
		}
	}

	if (unlikely(gc_state & KEY_GC_REAPING_DEAD_3)) {
		kdebug("dead wake");
		smp_mb();
		clear_bit(KEY_GC_REAPING_KEYTYPE, kgr_key_gc_flags);
		wake_up_bit(kgr_key_gc_flags, KEY_GC_REAPING_KEYTYPE);
	}

	if (gc_state & KEY_GC_REAP_AGAIN)
		schedule_work(kgr_key_gc_work);
	kleave(" [end %x]", gc_state);
	return;

	/* We found an unreferenced key - once we've removed it from the tree,
	 * we can safely drop the lock.
	 */
found_unreferenced_key:
	kdebug("unrefd key %d", key->serial);
	rb_erase(&key->serial_node, kgr_key_serial_tree);
	spin_unlock(kgr_key_serial_lock);

	list_add_tail(&key->graveyard_link, &graveyard);
	gc_state |= KEY_GC_REAP_AGAIN;
	goto maybe_resched;

	/* We found a keyring and we need to check the payload for links to
	 * dead or expired keys.  We don't flag another reap immediately as we
	 * have to wait for the old payload to be destroyed by RCU before we
	 * can reap the keys to which it refers.
	 */
found_keyring:
	spin_unlock(kgr_key_serial_lock);
	kgr_key_gc_keyring(key, limit);
	goto maybe_resched;

	/* We found a dead key that is still referenced.  Reset its type and
	 * destroy its payload with its semaphore held.
	 */
destroy_dead_key:
	spin_unlock(kgr_key_serial_lock);
	kdebug("destroy key %d", key->serial);
	down_write(&key->sem);
	key->type = kgr_key_type_dead;
	if ((*kgr_key_gc_dead_keytype)->destroy)
		(*kgr_key_gc_dead_keytype)->destroy(key);
	memset(&key->payload, KEY_DESTROY, sizeof(key->payload));
	up_write(&key->sem);
	goto maybe_resched;
}

static int kgr_patch_cve_2015_7872_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_cve_2015_7872_init(void)
{
	return kgr_patch_cve_2015_7872_kallsyms();
}
